/*******************************************************************************
 * Copyright 2003 Seapine Software, Inc. All rights reserved.
 * 
 * IMPORTANT:  This Software is supplied to you by Seapine Software, Inc. (Seapine)
 * in consideration of your agreement to the following terms, and your use, installation,
 * modification or redistribution of this Software constitutes acceptance of these terms.
 * If you do not agree with these terms, please do not use, install, modify or redistribute
 * this Software.
 * 
 * In consideration of your agreement to abide by the following terms, and subject to these
 * terms, Seapine grants you a personal, non-exclusive license, under Seapine copyrights in
 * this original Seapine software (the Software), to use, reproduce, modify and redistribute
 * the Software, with or without modifications, in source and/or binary forms; provided that
 * if you redistribute the Software in its entirety and without modifications, you must retain
 * this notice and the following text and disclaimers in all such redistributions of the Software.
 * Neither the name, trademarks, service marks or logos of Seapine, Inc. may be used to endorse
 * or promote products derived from the Software without specific prior written permission from
 * Seapine.  Except as expressly stated in this notice, no other rights or licenses, express or
 * implied, are granted by Seapine herein, including but not limited to any patent rights that
 * may be infringed by your derivative works or by other works in which the Software may be
 * incorporated.
 * 
 * The Software is provided by Seapine on an "AS IS" basis.  SEAPINE MAKES NO WARRANTIES,
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF NON-INFRINGEMENT,
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, REGARDING THE SOFTWARE OR ITS USE AND
 * OPERATION ALONE OR IN COMBINATION WITH YOUR PRODUCTS. 
 * 
 * IN NO EVENT SHALL SEAPINE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) ARISING IN ANY WAY OUT OF THE USE,
 * REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION OF THE SOFTWARE, HOWEVER CAUSED AND WHETHER
 * UNDER THEORY OF CONTRACT, TORT (INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN
 * IF SEAPINE HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <direct.h>  // for mkdir()
#include <errno.h>   // for ENOENT
#include <fcntl.h>   // following 4 are for _open() file locking
#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>

#include <windows.h> // for CopyFile()

//////////////////////////////////////////////////////////////////////////////

/*****
 Before compiling this trigger script, you will want to configure each of the
 variables as described below.  Many defaults may be acceptable.
   sSSCMExecutable - The full path to the sscm.exe executable
   sErrorLog - A full path to an error log (or NULL for no error log).  This
               log holds all of the errors during the sync process.
   sActivityLog - A full path to an activity log (or NULL for no activity
                  log).  This log holds all of the output of the sscm.exe
                  command line activity.
   sLockFile - A full path to a lock file used to keep the sscm.exe client
               from blocking on the activity log.  This can be NULL if the
               sActivityLog is NULL.
   sLocalServer - The IP:port of the local SCM server to sync from.
   sLocalUser - The username to login to the local SCM server as.
   sLocalUserPassword - The password for the above local username.
   sLocalSyncBranch - The local branch to sync.
   sLocalWorkingDir - The working dir for the topmost repository in the
                      local branch.
   sRemoteServer - The IP:port of the remote SCM server to sync to.
   sRemoteUser - The username to login to the remote SCM server as.
   sRemoteUserPassword - The password for the above remote username.
   sRemoteSyncBranch - The remote branch to sync to.
   sRemoteWorkingDir - The local working directory for the topmost repository
                       in the remote branch.
   sTempDir - A location that can be used to store temporary files.  If this
              variable is NULL, a path will be made off of the directory that
              this script is run out of.
*****/

const char sSSCMExecutable[] = "\"C:\\Program Files\\Seapine\\Surround SCM\\sscm.exe\"";
const char *sErrorLog = "C:\\scmsync\\SyncErrors.log";
const char *sActivityLog = "C:\\scmsync\\SyncActivity.log";
const char *sLockFile = "C:\\scmsync\\SyncSCM.lck";

const char sLocalServer[] = "192.168.1.100:4900";
const char sLocalUser[] = "scmsync";
const char sLocalUserPassword[] = "scmsync";
const char sLocalSyncBranch[] = "mainline";
const char sLocalWorkingDir[] = "C:\\scmsync\\local";

const char sRemoteServer[] = "192.168.1.101:4900";
const char sRemoteUser[] = "scmsync";
const char sRemoteUserPassword[] = "scmsync";
const char sRemoteSyncBranch[] = "mainline";
const char sRemoteWorkingDir[] = "C:\\scmsync\\remote";

const char *sTempDir = "C:\\scmsync\\temp";

/*****
 Before using this script, you will want to do the following:
 1. Compile this script.  This code was built with Visual C++ 6.0, but could
    be built by most standard Windows C++ compilers.  The command line used
    to compile this script was "cl.exe SyncSCM.cpp".
 2. Configure both the local server and the remote server to have an "scmsync"
    user, or similar.  Make sure the user name and password match the ones
    that were entered in the script.  Also be sure that the local user has
    access to get files out of the specified SCM repositories and that the
    remote user can add, get, and check in files to their respective
    repositories.
 3. The local "scmsync" user should have a working directory set up for the
    topmost repository on the local branch.  This working directory should
    match the one entered in the script variable.  Be sure to set this working
    directory up from the server machine where the SyncSCM script will run, as
    working directories can differ between machines.
 4. The remote "scmsync" user should also have a local working directory.  Set
    it up in the same way as the above step.
 5. Be sure that the local and remote working directories exist, as well as
    the directories that the log files will reside in and the temporary
    directory.
 6. Setup a trigger to fire this script from Surround SCM.  The trigger should
    be set to only run in the specified local sync branch, at the topmost
    repository, on all "*" files.  It should fire "after" "(all)" events.  Be
    sure to put the full path to the SyncSCM script in the "Run this script"
    field.


 Notes:
 1. The remote server can be setup to synchronize back to the local server if
    it is setup the same way.  Just swtich the values of the local and remote
    variables and follow the same steps on the remote server.
 2. This script will only synchronize one branch from one server to another.
    At this time, it is not possible to synchronize multiple branches across
    multiple servers.
 3. This script will attempt to avoid checking in files which it suspects to
    be checked in at the same time by comparing the previous local version of
    a file to the remote current version of a file.  If a file gets out of
    sync in this way, it will become necessary to manually check out the local
    or remote file and update it with the changes made in the other location.
    Check the error log periodically to see if a manual check in is required.
    Typically, the two files can become out of sync if two users check in a
    file at nearly the same time (either both checked in locally, or one local
    file and one remote file).  Having keyword expansion on will also cause
    this diff to fail.
 4. Renaming a repository will cause this script to fail, as it will no longer
    know where to look for files.  Renaming files is synchronized however.
 5. A branch can be renamed as long as the script is changed to reflect the
    new branch name.
 6. If the remote server has a file exclusively checked out, the check in
    synchronization will fail, and an error written to the error log.  You
    should manually synchronize this file whenever it becomes possible.
 7. Comments are not preserved across the synchronization, although they could
    be.  Feel free to modify this script to include that functionality if it
    is needed.
 8. The "Attach to defect" event is not synchronized, by design.
 9. Security information is not synchronized between the two repositories.
    This would have to be done manually.
*****/

//////////////////////////////////////////////////////////////////////////////

// These variables are the ones retrieved from the environment that the
// SCM trigger sets for us.
const char *sTriggerDesc = NULL;
const char *sTriggerEvent = NULL;
const char *sTriggerBeforeAfter = NULL;
const char *sTriggerDate = NULL;
const char *sTriggerTime = NULL;
const char *sTriggerUser = NULL;
const char *sTriggerClientIP = NULL;
const char *sTriggerClientMachine = NULL;
const char *sTriggerMainline = NULL;
const char *sTriggerBranch = NULL;
const char *sTriggerBranchType = NULL;
const char *sTriggerRepository = NULL;
const char *sTriggerFile = NULL;
const char *sTriggerFileVersion = NULL;
const char *sTriggerLocalFile = NULL;
const char *sTriggerComment = NULL;
const char *sTriggerSourceBranch = NULL;
const char *sTriggerDefectNumber = NULL;
const char *sTriggerShareRepository = NULL;
const char *sTriggerOldFileName = NULL;
const char *sTriggerPermRemove = NULL;

//////////////////////////////////////////////////////////////////////////////

// These functions handle most of the internals of synchronization, check the
// function header and internal comments for more details.

int DoAddFile(FILE *fpErrorLog);
int DoCheckInFile(FILE *fpErrorLog);
int DoRemoveFile(FILE *fpErrorLog);
int DoRenameFile(FILE *fpErrorLog);
int DoRestoreFile(FILE *fpErrorLog);
int DoBreakShareFile(FILE *fpErrorLog);
int DoCreateShareFile(FILE *fpErrorLog);
int DoCommonGetToLocalFile(FILE *fpErrorLog);
int DoCommonGet(FILE *fpErrorLog, bool bGetFromLocal, const char *sToPath, const char *sVersion);
void DoCommonMkRepository(FILE *fpErrorLog);
int DoCommonAddFromRemoteFile(FILE *fpErrorLog, const char *sFromOp);

bool GetLockForCLI();
void ClearLockForCLI();
int fhLockFile = -1; // Handle to the lock file, the file is deleted when this is closed.

bool CopyFileAndMkdir(const char *sSourceFile, const char *sDestFile);
bool mkdir_recursive(const char *sPath);

bool BuildLocalFileWorkingPath();
bool BuildRemoteFileWorkingPath();

char *sLocalFileWorkingPath = NULL;  // Full path to the local working file
char *sRemoteFileWorkingPath = NULL; // Full path to the remote working file

bool AreTheseFilesIdentical(const char *sFile1, const char *sFile2);

//////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
   int nReturn = 0;
   FILE *fpErrorLog = NULL;
   if(sErrorLog)
   {
      if(!(fpErrorLog = fopen(sErrorLog, "a")))
         return -1;  // Error opening error log, and no way to report it, quit.
   }

   // Get all of Surround SCM's environment variables (some may go unused)
   sTriggerDesc = getenv("SSCM_TRIGGERDESC");
   sTriggerEvent = getenv("SSCM_EVENT");
   sTriggerBeforeAfter = getenv("SSCM_BEFOREAFTER");
   sTriggerDate = getenv("SSCM_DATE");
   sTriggerTime = getenv("SSCM_TIME");
   sTriggerUser = getenv("SSCM_USER");
   sTriggerClientIP = getenv("SSCM_CLIENTIP");
   sTriggerClientMachine = getenv("SSCM_CLIENTMACHINE");
   sTriggerMainline = getenv("SSCM_MAINLINE");
   sTriggerBranch = getenv("SSCM_BRANCH");
   sTriggerBranchType = getenv("SSCM_BRANCHTYPE");
   sTriggerRepository = getenv("SSCM_REPOSITORY");
   sTriggerFile = getenv("SSCM_FILE");
   sTriggerFileVersion = getenv("SSCM_FILEVERSION");
   sTriggerLocalFile = getenv("SSCM_LOCALFILE");
   sTriggerComment = getenv("SSCM_COMMENT");
   sTriggerSourceBranch = getenv("SSCM_SOURCEBRANCH");
   sTriggerDefectNumber = getenv("SSCM_DEFECTNUMBER");
   sTriggerShareRepository = getenv("SSCM_SHAREREPOSITORY");
   sTriggerOldFileName = getenv("SSCM_OLDFILENAME");
   sTriggerPermRemove = getenv("SSCM_PERMREMOVE");

   // Check to make sure we have the env vars that we need
   if(!sTriggerDate || !sTriggerTime)      
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "!! !! : One or more trigger environment variables was invalid.\n");
      nReturn = -1;
   }
   else if(!sTriggerEvent || !sTriggerUser || !sTriggerClientIP ||
           !sTriggerClientMachine || !sTriggerBranch ||
           !sTriggerRepository || !sTriggerFile)
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : One or more trigger environment variables was invalid.\n",
                             sTriggerDate, sTriggerTime);
      nReturn = -1;
   }
   else if(strcmp(sLocalSyncBranch, sTriggerBranch))
   {
      // Check to make sure we have been fired on the correct local branch.
      // Put an error in the log if we haven't.
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : The trigger fired on the branch %s, but is configured to sync the branch %s.\n",
                             sTriggerBranch, sLocalSyncBranch);
      nReturn = -1;
   }

   // Start a sync for the appropriate event
   if(nReturn == 0)
   {
      if(stricmp(sTriggerUser, sLocalUser) == 0)
         nReturn = 0;  // This is our sync user performing a sync operation,
                       // so don't try to copy this to the other server.
      else if(strcmp(sTriggerEvent, "Add") == 0)
         nReturn = DoAddFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Add from branch") == 0)
         nReturn = DoAddFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Attach to defect") == 0)
         nReturn = 0; // Nothing to do
      else if(strcmp(sTriggerEvent, "Check in") == 0)
         nReturn = DoCheckInFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Promote from") == 0)
         nReturn = DoCheckInFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Rebase") == 0)
         nReturn = DoCheckInFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Remove") == 0)
         nReturn = DoRemoveFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Rename file") == 0)
         nReturn = DoRenameFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Restore") == 0)
         nReturn = DoRestoreFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Share broken") == 0)
         nReturn = DoBreakShareFile(fpErrorLog);
      else if(strcmp(sTriggerEvent, "Share created") == 0)
         nReturn = DoCreateShareFile(fpErrorLog);
      else
      {
         if(fpErrorLog)
            fprintf(fpErrorLog, "%s %s : Trigger event does not match any known event type.\n",
                                sTriggerDate, sTriggerTime);
         nReturn = -1;  // Some unknown event, we should never hit this
      }
   }

   if(sLocalFileWorkingPath)
      delete [] sLocalFileWorkingPath;
   if(sRemoteFileWorkingPath)
      delete [] sRemoteFileWorkingPath;

   if(fpErrorLog)
      fclose(fpErrorLog);
   return nReturn;
}

//////////////////////////////////////////////////////////////////////////////

// Handles "Add" and "Add from branch"
int DoAddFile(FILE *fpErrorLog)
{
   char sCmdLine[4096] = { 0 };
   int nReturn = 0;

   // Get the file just added out of the local server
   nReturn = DoCommonGetToLocalFile(fpErrorLog);
   if(nReturn)
      return nReturn;

   // Setup our local and remote working file paths
   if(!BuildLocalFileWorkingPath() || !BuildRemoteFileWorkingPath())
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error getting local path information for %s/%s.\n",
                             sTriggerDate, sTriggerTime,
                             sTriggerRepository, sTriggerFile);
      return -1;
   }

   // Copy the local file to the remote working path
   if(!CopyFileAndMkdir(sLocalFileWorkingPath, sRemoteFileWorkingPath))
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error making working copy from %s to %s.\n",
                             sTriggerDate, sTriggerTime,
                             sLocalFileWorkingPath, sRemoteFileWorkingPath);
      return -1;
   }

   // Create our target repository in case its needed (often it won't be)
   DoCommonMkRepository(fpErrorLog);

   // Add the file from the remote path
   nReturn = DoCommonAddFromRemoteFile(fpErrorLog, "SCMSyncAdd");

   return nReturn;
}

//////////////////////////////////////////////////////////////////////////////

// Handles "Check in", "Promote from", and "Rebase"
int DoCheckInFile(FILE *fpErrorLog)
{
   char sCmdLine[4096] = { 0 };
   int nReturn = 0;

   // Be sure we have some version information from the trigger
   if(sTriggerFileVersion && sTriggerFileVersion[0] != 0)
   {
      int nFileVersion = atoi(sTriggerFileVersion);
      if(nFileVersion > 1)
      {
         // Normal check out --> check in, so verify that we're checking in the right file
         char sTempDirLocal[1024] = { 0 };
         char sTempDirRemote[1024] = { 0 };
         char sTempFileLocal[1024] = { 0 };
         char sTempFileRemote[1024] = { 0 };

         // Setup our temporary directories for the files we will diff
         if(sTempDir)
         {
            strcpy(sTempDirLocal, sTempDir);
            strcat(sTempDirLocal, "\\");
            strcpy(sTempDirRemote, sTempDir);
            strcat(sTempDirRemote, "\\");
         }
         strcat(sTempDirLocal, "trltmpXXXXXX");
         strcat(sTempDirRemote, "trrtmpXXXXXX");
         _mktemp(sTempDirLocal);
         _mktemp(sTempDirRemote);
         strcpy(sTempFileLocal, sTempDirLocal);
         strcpy(sTempFileRemote, sTempDirRemote);
         strcat(sTempFileLocal, "\\");
         strcat(sTempFileRemote, "\\");
         strcat(sTempFileLocal, sTriggerFile);
         strcat(sTempFileRemote, sTriggerFile);

         // Create the directories
         if(mkdir_recursive(sTempDirLocal)  &&  mkdir_recursive(sTempDirRemote))
         {
            char sLocalPrevVer[16];
            sprintf(sLocalPrevVer, "%u", nFileVersion - 1);
            nReturn = DoCommonGet(fpErrorLog, true, sTempDirLocal, sLocalPrevVer);
            if(!nReturn)
            {
               nReturn = DoCommonGet(fpErrorLog, false, sTempDirRemote, NULL);
               if(!nReturn)
               {
                  // We have the two files (local previous version, remote current
                  //  version).  Ensure that they are the same.
                  if(!AreTheseFilesIdentical(sTempFileLocal, sTempFileRemote))
                  {
                     // They're different, log an error.
                     nReturn = -1;
                     if(fpErrorLog)
                        fprintf(fpErrorLog, "%s %s : Error performing checkin operation on %s/%s, previous local version differs from current remote version.  Performing a manual check in is necessary.\n",
                                            sTriggerDate, sTriggerTime,
                                            sTriggerRepository, sTriggerFile);
                  }
                  else
                     nReturn = 0;

                  remove(sTempFileRemote);
               }

               remove(sTempFileLocal);
            }

            _rmdir(sTempDirLocal);
            _rmdir(sTempDirRemote);
         }
         else // bad mkdir
         {
            _rmdir(sTempDirLocal);
            _rmdir(sTempDirRemote);
            nReturn = -1;
            if(fpErrorLog)
               fprintf(fpErrorLog, "%s %s : Error performing checkin operation on %s/%s, Could not create directories for performing diff.\n",
                                   sTriggerDate, sTriggerTime,
                                   sTriggerRepository, sTriggerFile);
         }
      }
      else
      {
         nReturn = 0;
      }

      if(nReturn == 0) // All is well, do the check in (or possibly add)
      {
         // Get the file just checked in from the local server
         nReturn = DoCommonGetToLocalFile(fpErrorLog);
         if(nReturn)
            return nReturn;

         if(!BuildLocalFileWorkingPath() || !BuildRemoteFileWorkingPath())
         {
            if(fpErrorLog)
               fprintf(fpErrorLog, "%s %s : Error getting local path information for %s/%s.\n",
                                   sTriggerDate, sTriggerTime,
                                   sTriggerRepository, sTriggerFile);
            return -1;
         }

         if(nFileVersion > 1) // normal file version, so do a check out, check in
         {
            // Check out the file on the remote machine
            strcpy(sCmdLine, sSSCMExecutable);
            strcat(sCmdLine, " checkout ");
            strcat(sCmdLine, sTriggerFile);
            strcat(sCmdLine, " -b");
            strcat(sCmdLine, sRemoteSyncBranch);
            strcat(sCmdLine, " -p");
            strcat(sCmdLine, sTriggerRepository);
            strcat(sCmdLine, " -ccSCMSyncCheckIn -wskip -y");
            strcat(sCmdLine, sRemoteUser);
            strcat(sCmdLine, ":");
            strcat(sCmdLine, sRemoteUserPassword);
            strcat(sCmdLine, " -z");
            strcat(sCmdLine, sRemoteServer);
            if(sActivityLog)
            {
               strcat(sCmdLine, " >> ");
               strcat(sCmdLine, sActivityLog);
               if(GetLockForCLI())
               {
                  nReturn = system(sCmdLine);
                  ClearLockForCLI();
               }
               else
               {
                  nReturn = -1;
                  if(fpErrorLog)
                     fprintf(fpErrorLog, "%s %s : Error obtaining lock file for CLI.\n",
                                         sTriggerDate, sTriggerTime);
               }
            }
            else
            {
               nReturn = system(sCmdLine);
            }

            if(nReturn)
            {
               if(fpErrorLog)
                  fprintf(fpErrorLog, "%s %s : Error performing checkout operation on %s/%s.\n",
                                      sTriggerDate, sTriggerTime,
                                      sTriggerRepository, sTriggerFile);
               return nReturn;
            }

            // Copy our file to the checked out file path
            if(!CopyFileAndMkdir(sLocalFileWorkingPath, sRemoteFileWorkingPath))
            {
               if(fpErrorLog)
                  fprintf(fpErrorLog, "%s %s : Error making working copy from %s to %s.\n",
                                      sTriggerDate, sTriggerTime,
                                      sLocalFileWorkingPath, sRemoteFileWorkingPath);
               return -1;
            }

            // Check in the file on the remote machine
            strcpy(sCmdLine, sSSCMExecutable);
            strcat(sCmdLine, " checkin ");
            strcat(sCmdLine, sTriggerFile);
            strcat(sCmdLine, " -b");
            strcat(sCmdLine, sRemoteSyncBranch);
            strcat(sCmdLine, " -p");
            strcat(sCmdLine, sTriggerRepository);
            strcat(sCmdLine, " -ccSCMSyncCheckIn -g- -u -y");
            strcat(sCmdLine, sRemoteUser);
            strcat(sCmdLine, ":");
            strcat(sCmdLine, sRemoteUserPassword);
            strcat(sCmdLine, " -z");
            strcat(sCmdLine, sRemoteServer);
            if(sActivityLog)
            {
               strcat(sCmdLine, " >> ");
               strcat(sCmdLine, sActivityLog);
               if(GetLockForCLI())
               {
                  nReturn = system(sCmdLine);
                  ClearLockForCLI();
               }
               else
               {
                  nReturn = -1;
                  if(fpErrorLog)
                     fprintf(fpErrorLog, "%s %s : Error obtaining lock file for CLI.\n",
                                         sTriggerDate, sTriggerTime);
               }
            }
            else
            {
               nReturn = system(sCmdLine);
            }

            if(nReturn)
            {
               if(fpErrorLog)
                  fprintf(fpErrorLog, "%s %s : Error performing checkin operation on %s/%s.\n",
                                      sTriggerDate, sTriggerTime,
                                      sTriggerRepository, sTriggerFile);
               return nReturn;
            }
         }
         else // nFileVersion <= 1
         {
            // In this case, we're most likely a promote/rebase of a new file, so add it

            // Copy our file to the checked out file path
            if(!CopyFileAndMkdir(sLocalFileWorkingPath, sRemoteFileWorkingPath))
            {
               if(fpErrorLog)
                  fprintf(fpErrorLog, "%s %s : Error making working copy from %s to %s.\n",
                                      sTriggerDate, sTriggerTime,
                                      sLocalFileWorkingPath, sRemoteFileWorkingPath);
               return -1;
            }

            // Create our target repository in case its needed
            DoCommonMkRepository(fpErrorLog);

            // Add the file from the remote path
            nReturn = DoCommonAddFromRemoteFile(fpErrorLog, "SCMSyncCheckIn");

            if(nReturn)
            {
               if(fpErrorLog)
                  fprintf(fpErrorLog, "%s %s : Error performing add operation on %s/%s.\n",
                                      sTriggerDate, sTriggerTime,
                                      sTriggerRepository, sTriggerFile);
               return nReturn;
            }
         }
      }// nReturn == 0 after diff
   }// valid file version with trigger
   else
   {
      nReturn = -1;
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error performing checkin operation on %s/%s, no version information available.  Performing a manual check in is recommended.\n",
                             sTriggerDate, sTriggerTime,
                             sTriggerRepository, sTriggerFile);
      return nReturn;
   }

   return nReturn;
}

//////////////////////////////////////////////////////////////////////////////

// Handles "Remove"
int DoRemoveFile(FILE *fpErrorLog)
{
   char sCmdLine[4096] = { 0 };
   int nReturn = 0;

   // Check for a permanent remove
   bool bPermRemove = false;
   if(strcmp(sTriggerPermRemove, "yes") == 0)
      bPermRemove = true;

   // Remove the file on the remote machine
   strcpy(sCmdLine, sSSCMExecutable);
   strcat(sCmdLine, " rm ");
   strcat(sCmdLine, sTriggerFile);
   strcat(sCmdLine, " -b");
   strcat(sCmdLine, sRemoteSyncBranch);
   strcat(sCmdLine, " -p");
   strcat(sCmdLine, sTriggerRepository);
   if(bPermRemove)
      strcat(sCmdLine, " -d -q");
   strcat(sCmdLine, " -ccSCMSyncRemove -y");
   strcat(sCmdLine, sRemoteUser);
   strcat(sCmdLine, ":");
   strcat(sCmdLine, sRemoteUserPassword);
   strcat(sCmdLine, " -z");
   strcat(sCmdLine, sRemoteServer);
   if(sActivityLog)
   {
      strcat(sCmdLine, " >> ");
      strcat(sCmdLine, sActivityLog);
      if(GetLockForCLI())
      {
         nReturn = system(sCmdLine);
         ClearLockForCLI();
      }
      else
      {
         nReturn = -1;
         if(fpErrorLog)
            fprintf(fpErrorLog, "%s %s : Error obtaining lock file for CLI.\n",
                                sTriggerDate, sTriggerTime);
      }
   }
   else
   {
      nReturn = system(sCmdLine);
   }

   if(nReturn)
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error performing remove operation on %s/%s.\n",
                             sTriggerDate, sTriggerTime,
                             sTriggerRepository, sTriggerFile);
      return nReturn;
   }

   return nReturn;
}

//////////////////////////////////////////////////////////////////////////////

// Handles "Rename file"
int DoRenameFile(FILE *fpErrorLog)
{
   char sCmdLine[4096] = { 0 };
   int nReturn = 0;

   // Rename the file on the remote machine
   strcpy(sCmdLine, sSSCMExecutable);
   strcat(sCmdLine, " rename ");
   strcat(sCmdLine, sTriggerOldFileName);
   strcat(sCmdLine, " ");
   strcat(sCmdLine, sTriggerFile);
   strcat(sCmdLine, " -b");
   strcat(sCmdLine, sRemoteSyncBranch);
   strcat(sCmdLine, " -p");
   strcat(sCmdLine, sTriggerRepository);
   strcat(sCmdLine, " -ccSCMSyncRename -y");
   strcat(sCmdLine, sRemoteUser);
   strcat(sCmdLine, ":");
   strcat(sCmdLine, sRemoteUserPassword);
   strcat(sCmdLine, " -z");
   strcat(sCmdLine, sRemoteServer);
   if(sActivityLog)
   {
      strcat(sCmdLine, " >> ");
      strcat(sCmdLine, sActivityLog);
      if(GetLockForCLI())
      {
         nReturn = system(sCmdLine);
         ClearLockForCLI();
      }
      else
      {
         nReturn = -1;
         if(fpErrorLog)
            fprintf(fpErrorLog, "%s %s : Error obtaining lock file for CLI.\n",
                                sTriggerDate, sTriggerTime);
      }
   }
   else
   {
      nReturn = system(sCmdLine);
   }

   if(nReturn)
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error performing rename operation on %s/%s.\n",
                             sTriggerDate, sTriggerTime,
                             sTriggerRepository, sTriggerOldFileName);
      return nReturn;
   }

   return nReturn;
}

//////////////////////////////////////////////////////////////////////////////

// Handles "Restore"
int DoRestoreFile(FILE *fpErrorLog)
{
   char sCmdLine[4096] = { 0 };
   int nReturn = 0;

   // Restore the file on the remote machine
   strcpy(sCmdLine, sSSCMExecutable);
   strcat(sCmdLine, " restore ");
   strcat(sCmdLine, sTriggerFile);
   strcat(sCmdLine, " -b");
   strcat(sCmdLine, sRemoteSyncBranch);
   strcat(sCmdLine, " -p");
   strcat(sCmdLine, sTriggerRepository);
   strcat(sCmdLine, " -ccSCMSyncRestore -y");
   strcat(sCmdLine, sRemoteUser);
   strcat(sCmdLine, ":");
   strcat(sCmdLine, sRemoteUserPassword);
   strcat(sCmdLine, " -z");
   strcat(sCmdLine, sRemoteServer);
   if(sActivityLog)
   {
      strcat(sCmdLine, " >> ");
      strcat(sCmdLine, sActivityLog);
      if(GetLockForCLI())
      {
         nReturn = system(sCmdLine);
         ClearLockForCLI();
      }
      else
      {
         nReturn = -1;
         if(fpErrorLog)
            fprintf(fpErrorLog, "%s %s : Error obtaining lock file for CLI.\n",
                                sTriggerDate, sTriggerTime);
      }
   }
   else
   {
      nReturn = system(sCmdLine);
   }

   if(nReturn)
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error performing restore operation on %s/%s.\n",
                             sTriggerDate, sTriggerTime,
                             sTriggerRepository, sTriggerFile);
      return nReturn;
   }

   return nReturn;
}

//////////////////////////////////////////////////////////////////////////////

// Handles "Share broken"
int DoBreakShareFile(FILE *fpErrorLog)
{
   char sCmdLine[4096] = { 0 };
   int nReturn = 0;

   // Break the share on the remote machine
   strcpy(sCmdLine, sSSCMExecutable);
   strcat(sCmdLine, " breakshare ");
   strcat(sCmdLine, sTriggerFile);
   strcat(sCmdLine, " -b");
   strcat(sCmdLine, sRemoteSyncBranch);
   strcat(sCmdLine, " -p");
   strcat(sCmdLine, sTriggerRepository);
   strcat(sCmdLine, " -ccSCMSyncBreakShare -y");
   strcat(sCmdLine, sRemoteUser);
   strcat(sCmdLine, ":");
   strcat(sCmdLine, sRemoteUserPassword);
   strcat(sCmdLine, " -z");
   strcat(sCmdLine, sRemoteServer);
   if(sActivityLog)
   {
      strcat(sCmdLine, " >> ");
      strcat(sCmdLine, sActivityLog);
      if(GetLockForCLI())
      {
         nReturn = system(sCmdLine);
         ClearLockForCLI();
      }
      else
      {
         nReturn = -1;
         if(fpErrorLog)
            fprintf(fpErrorLog, "%s %s : Error obtaining lock file for CLI.\n",
                                sTriggerDate, sTriggerTime);
      }
   }
   else
   {
      nReturn = system(sCmdLine);
   }

   if(nReturn)
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error performing breakshare operation on %s/%s.\n",
                             sTriggerDate, sTriggerTime,
                             sTriggerRepository, sTriggerFile);
      return nReturn;
   }

   return nReturn;
}

//////////////////////////////////////////////////////////////////////////////

// Handles "Share created"
int DoCreateShareFile(FILE *fpErrorLog)
{
   char sCmdLine[4096] = { 0 };
   int nReturn = 0;

   // Create our target repository in case its needed
   DoCommonMkRepository(fpErrorLog);

   // Create the same share of the file on the remote server
   strcpy(sCmdLine, sSSCMExecutable);
   strcat(sCmdLine, " share ");
   strcat(sCmdLine, sTriggerShareRepository);
   strcat(sCmdLine, " ");
   strcat(sCmdLine, sTriggerFile);
   strcat(sCmdLine, " -b");
   strcat(sCmdLine, sRemoteSyncBranch);
   strcat(sCmdLine, " -p");
   strcat(sCmdLine, sTriggerRepository);
   strcat(sCmdLine, " -ccSCMSyncShare -y");
   strcat(sCmdLine, sRemoteUser);
   strcat(sCmdLine, ":");
   strcat(sCmdLine, sRemoteUserPassword);
   strcat(sCmdLine, " -z");
   strcat(sCmdLine, sRemoteServer);
   if(sActivityLog)
   {
      strcat(sCmdLine, " >> ");
      strcat(sCmdLine, sActivityLog);
      if(GetLockForCLI())
      {
         nReturn = system(sCmdLine);
         ClearLockForCLI();
      }
      else
      {
         nReturn = -1;
         if(fpErrorLog)
            fprintf(fpErrorLog, "%s %s : Error obtaining lock file for CLI.\n",
                                sTriggerDate, sTriggerTime);
      }
   }
   else
   {
      nReturn = system(sCmdLine);
   }

   if(nReturn)
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error performing share operation on %s/%s.\n",
                             sTriggerDate, sTriggerTime,
                             sTriggerRepository, sTriggerFile);
      return nReturn;
   }

   return nReturn;
}

//////////////////////////////////////////////////////////////////////////////

// Get the local file to the local working directory
int DoCommonGetToLocalFile(FILE *fpErrorLog)
{
   // Get the file just checked in/added from the local server
   return DoCommonGet(fpErrorLog, true, NULL, NULL);
}

//////////////////////////////////////////////////////////////////////////////

int DoCommonGet(FILE *fpErrorLog, bool bGetFromLocal, const char *sToPath, const char *sVersion)
{
   char sCmdLine[4096] = { 0 };
   int nReturn = 0;

   // Get the file off of a specified server
   strcpy(sCmdLine, sSSCMExecutable);
   strcat(sCmdLine, " get ");
   strcat(sCmdLine, sTriggerFile);
   strcat(sCmdLine, " -b");
   if(bGetFromLocal)
      strcat(sCmdLine, sTriggerBranch);
   else
      strcat(sCmdLine, sRemoteSyncBranch);
   strcat(sCmdLine, " -p");
   strcat(sCmdLine, sTriggerRepository);
   if(sToPath)
   {
      strcat(sCmdLine, " -d");
      strcat(sCmdLine, sToPath);
   }
   if(sVersion)
   {
      strcat(sCmdLine, " -v");
      strcat(sCmdLine, sVersion);
   }
   strcat(sCmdLine, " -f -e -wreplace -y");
   if(bGetFromLocal)
   {
      strcat(sCmdLine, sLocalUser);
      strcat(sCmdLine, ":");
      strcat(sCmdLine, sLocalUserPassword);
      strcat(sCmdLine, " -z");
      strcat(sCmdLine, sLocalServer);
   }
   else
   {
      strcat(sCmdLine, sRemoteUser);
      strcat(sCmdLine, ":");
      strcat(sCmdLine, sRemoteUserPassword);
      strcat(sCmdLine, " -z");
      strcat(sCmdLine, sRemoteServer);
   }
   if(sActivityLog)
   {
      strcat(sCmdLine, " >> ");
      strcat(sCmdLine, sActivityLog);
      if(GetLockForCLI())
      {
         nReturn = system(sCmdLine);
         ClearLockForCLI();
      }
      else
      {
         nReturn = -1;
         if(fpErrorLog)
            fprintf(fpErrorLog, "%s %s : Error obtaining lock file for CLI.\n",
                                sTriggerDate, sTriggerTime);
      }
   }
   else
   {
      nReturn = system(sCmdLine);
   }

   if(nReturn)
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error performing %s get operation on %s/%s %s.\n",
                             sTriggerDate, sTriggerTime,
                             bGetFromLocal ? "local" : "remote",
                             sTriggerRepository, sTriggerFile,
                             sVersion ? sVersion : "");
   }

   return nReturn;
}

//////////////////////////////////////////////////////////////////////////////

// This function is called whenever we may need to make a repository for a new
// file.  In most cases the repository will already exist, and we will
// just silently fail.
void DoCommonMkRepository(FILE *fpErrorLog)
{
   // Be sure that the repository exists for the destination file (recursively)
   char sCmdLine[4096] = { 0 };

   char *sMkRepository = new char[strlen(sTriggerRepository)+1];
   if(sTriggerRepository[0] == '/')
      strcpy(sMkRepository, &sTriggerRepository[1]);
   else
      strcpy(sMkRepository, sTriggerRepository);

   char *a = sMkRepository;
   char *b = sMkRepository;
   bool bLastOne = false;
   while(*a && *a != '/')
      a++;
   while(*a && !bLastOne)
   {
      b = a+1;
      while(*b && *b != '/')
         b++;
      if(!*b)
         bLastOne = true;
      *a = 0;
      *b = 0;
      strcpy(sCmdLine, sSSCMExecutable);
      strcat(sCmdLine, " mkrepository ");
      strcat(sCmdLine, (a+1)); // repo to make
      strcat(sCmdLine, " -b");
      strcat(sCmdLine, sRemoteSyncBranch);
      strcat(sCmdLine, " -p"); // -p = parent repo
      strcat(sCmdLine, sMkRepository);
      strcat(sCmdLine, " -ccSCMSyncAdd -y");
      strcat(sCmdLine, sRemoteUser);
      strcat(sCmdLine, ":");
      strcat(sCmdLine, sRemoteUserPassword);
      strcat(sCmdLine, " -z");
      strcat(sCmdLine, sRemoteServer);
      if(sActivityLog)
      {
         strcat(sCmdLine, " >> ");
         strcat(sCmdLine, sActivityLog);
         if(GetLockForCLI())
         {
            system(sCmdLine);
            ClearLockForCLI();
         }
         else
         {
            if(fpErrorLog)
               fprintf(fpErrorLog, "%s %s : Error obtaining lock file for CLI.\n",
                                   sTriggerDate, sTriggerTime);
         }
      }
      else
      {
         system(sCmdLine);
      }

      *a = '/';
      if(!bLastOne)
         *b = '/';
      a = b;
   }
   delete [] sMkRepository;
}

//////////////////////////////////////////////////////////////////////////////

// Add the file from the remote path
int DoCommonAddFromRemoteFile(FILE *fpErrorLog, const char *sFromOp)
{
   // Add the file we retrieved to the remote server
   int nReturn;
   char sCmdLine[4096] = { 0 };

   strcpy(sCmdLine, sSSCMExecutable);
   strcat(sCmdLine, " add ");
   strcat(sCmdLine, sRemoteFileWorkingPath);
   strcat(sCmdLine, " -b");
   strcat(sCmdLine, sRemoteSyncBranch);
   strcat(sCmdLine, " -p");
   strcat(sCmdLine, sTriggerRepository);
   strcat(sCmdLine, " -cc");
   strcat(sCmdLine, sFromOp);
   strcat(sCmdLine, " -g- -y");
   strcat(sCmdLine, sRemoteUser);
   strcat(sCmdLine, ":");
   strcat(sCmdLine, sRemoteUserPassword);
   strcat(sCmdLine, " -z");
   strcat(sCmdLine, sRemoteServer);
   if(sActivityLog)
   {
      strcat(sCmdLine, " >> ");
      strcat(sCmdLine, sActivityLog);
      if(GetLockForCLI())
      {
         nReturn = system(sCmdLine);
         ClearLockForCLI();
      }
      else
      {
         nReturn = -1;
         if(fpErrorLog)
            fprintf(fpErrorLog, "%s %s : Error obtaining lock file for CLI.\n",
                                sTriggerDate, sTriggerTime);
      }
   }
   else
   {
      nReturn = system(sCmdLine);
   }

   if(nReturn)
   {
      if(fpErrorLog)
         fprintf(fpErrorLog, "%s %s : Error performing add operation on %s/%s.\n",
                             sTriggerDate, sTriggerTime,
                             sTriggerRepository, sTriggerFile);
      return nReturn;
   }

   return nReturn;
}

//////////////////////////////////////////////////////////////////////////////

// This function copies a file from source to destination, ensuring that the
// directories exist for the files.
bool CopyFileAndMkdir(const char *sSourceFile, const char *sDestFile)
{
   int pathlen = strlen(sDestFile);
   char *parent = (char *)malloc(pathlen + 1);
   strcpy(parent, sDestFile);
   if(parent[pathlen - 1] == '\\') parent[pathlen - 1] = '\0';
   char *end = strrchr(parent,'\\');
   if(end != NULL)
   {
      end[1] = '\0';
      mkdir_recursive(parent);
      return (CopyFile(sSourceFile, sDestFile, false) != 0);
   }
   else
      return false;
}

//////////////////////////////////////////////////////////////////////////////

// This function will create a directory and any parent directories that are
// needed for the child to exist.
bool mkdir_recursive(const char *sPath)
{
   // start by creating the directory and see if it works
   if(_mkdir(sPath)) {
      if(errno == ENOENT) {
         // parent doesn't exist: create it
         int pathlen = strlen(sPath);

         char *parent = (char *)malloc(pathlen + 1);
         strcpy(parent, sPath);
         // strip trailing directory seperator
         if(parent[pathlen - 1] == '\\') parent[pathlen - 1] = '\0';
         char *end = strrchr(parent,'\\');
         if(end != NULL) {
            end[1] = '\0'; // null terminate the parent directory name
         
            if(mkdir_recursive(parent)) {
               free (parent);
               // successfully made parent, now make this directory again
               return (mkdir_recursive(sPath));
            }
         }
         free (parent);
      } // errno == ENOENT
      return false;             // mkdir failed for who knows why
   }

   return true; // return true on success
}

//////////////////////////////////////////////////////////////////////////////

// This function will create a working path to the local file based on the
// working directory, file's repository, and file name.
bool BuildLocalFileWorkingPath()
{
   if(sLocalFileWorkingPath)
      return false; // Already have it set

   sLocalFileWorkingPath = new char[strlen(sLocalWorkingDir) + strlen(sTriggerRepository) + strlen(sTriggerFile)];
   if(!sLocalFileWorkingPath)
      return false; // Memory allocation error

   strcpy(sLocalFileWorkingPath, sLocalWorkingDir);
   char *a = (char *)sTriggerRepository + 1;
   while(*a  &&  *a != '/')
      a++;
   if(*a && *++a)
   {
      strcat(sLocalFileWorkingPath, "\\");
      strcat(sLocalFileWorkingPath, a);
      if(sLocalFileWorkingPath[strlen(sLocalFileWorkingPath)-1] == '/')
         sLocalFileWorkingPath[strlen(sLocalFileWorkingPath)-1] = 0;
   }
   for(a=sLocalFileWorkingPath; *a; a++)
      if(*a == '/')
         *a = '\\';
   strcat(sLocalFileWorkingPath, "\\");
   strcat(sLocalFileWorkingPath, sTriggerFile);
   return true;
}

//////////////////////////////////////////////////////////////////////////////

// This function will create a working path to the remote file based on the
// working directory, file's repository, and file name.
bool BuildRemoteFileWorkingPath()
{
   if(sRemoteFileWorkingPath)
      return false; // Already have it set

   sRemoteFileWorkingPath = new char[strlen(sRemoteWorkingDir) + strlen(sTriggerRepository) + strlen(sTriggerFile)];
   if(!sRemoteFileWorkingPath)
      return false; // Memory allocation error

   strcpy(sRemoteFileWorkingPath, sRemoteWorkingDir);
   char *a = (char *)sTriggerRepository + 1;
   while(*a  &&  *a != '/')
      a++;
   if(*a && *++a)
   {
      strcat(sRemoteFileWorkingPath, "\\");
      strcat(sRemoteFileWorkingPath, a);
      if(sRemoteFileWorkingPath[strlen(sRemoteFileWorkingPath)-1] == '/')
         sRemoteFileWorkingPath[strlen(sRemoteFileWorkingPath)-1] = 0;
   }
   for(a=sRemoteFileWorkingPath; *a; a++)
      if(*a == '/')
         *a = '\\';
   strcat(sRemoteFileWorkingPath, "\\");
   strcat(sRemoteFileWorkingPath, sTriggerFile);
   return true;
}

//////////////////////////////////////////////////////////////////////////////

// This function will wait up to 30 seconds to get a lock on the lock file.
// It returns false if it could not get a lock.
bool GetLockForCLI()
{
   bool bReturn = false;
   int nSecsToKeepTrying = 30000;  // Wait up to 30.000 seconds for a lock
   if(sLockFile)
   {
      while(nSecsToKeepTrying > 0)
      {
         fhLockFile = _open(sLockFile, _O_CREAT | _O_EXCL | _O_TEMPORARY, _S_IWRITE);
         if(fhLockFile == -1  &&  errno == EEXIST)
         {
            // Lock file is in use by someone else... wait for it.
            nSecsToKeepTrying -= 100;
            Sleep(100);
         }
         else if(fhLockFile != -1)
         {
            // Lock file is good
            bReturn = true;
            nSecsToKeepTrying = 0;
         }
         else
         {
            // Some error with the lock file, just quit
            bReturn = false;
            nSecsToKeepTrying = 0;
         }
      }
   }

   return bReturn;
}

//////////////////////////////////////////////////////////////////////////////

// This function clears the lock by closing the file handle (which deletes
// the file).
void ClearLockForCLI()
{
   if(fhLockFile != -1)
      _close(fhLockFile);
}

//////////////////////////////////////////////////////////////////////////////

// Does a full binary comparison of 2 files to ensure that they are the same
bool AreTheseFilesIdentical(const char *sFile1, const char *sFile2)
{
   if(!sFile1  ||  sFile1[0] == 0)
      return false;
   if(!sFile2  ||  sFile2[0] == 0)
      return false;

   FILE *fp1 = NULL;
   FILE *fp2 = NULL;

   fp1 = fopen(sFile1, "rb");
   fp2 = fopen(sFile2, "rb");

   if(!fp1 || !fp2)
   {
      if(fp1)
         fclose(fp1);
      if(fp2)
         fclose(fp2);
      return false;
   }

   char data1[1024];
   char data2[1024];
   unsigned long len1, len2;

   do
   {
      len1 = fread(data1, 1, 1024, fp1);
      len2 = fread(data2, 1, 1024, fp2);
      if(len1 != len2)
      {
         fclose(fp1);
         fclose(fp2);
         return false;
      }
      else if(len1 > 0)
      {
         if(memcmp(data1, data2, len1) != 0)
         {
            fclose(fp1);
            fclose(fp2);
            return false;
         }
      }
   } while(len1 > 0  &&  len2 > 0);

   fclose(fp1);
   fclose(fp2);
   return true;
}

//////////////////////////////////////////////////////////////////////////////
